/*
 * Copyright (c) 2010,2017 Gražvydas Ignotas
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <linux/input.h>
#include <linux/fb.h>

#include "font.c"

#define array_size(x) (sizeof(x) / sizeof(x[0]))

#define VALIDATION

#define NUB_RANGE_REQ 250

#define COL_DEF    0x7f7f7f
#define COL_TESTED 0xe8e840
#define COL_ACTIVE 0x00ff00

typedef struct {
	int x, y;
	int code;
	unsigned int active:1;
	unsigned int tested:1;
	const char *name;
} key_item;

static key_item key_items[] =
{
	{ 164, 192, KEY_UP,           0, 0, "U" },
	{  96, 256, KEY_LEFT,         0, 0, "L" },
	{ 222, 256, KEY_RIGHT,        0, 0, "R" },
	{ 164, 324, KEY_DOWN,         0, 0, "D" },
	{1036, 234, KEY_HOME,         0, 0, "A" },
	{1196, 234, KEY_END,          0, 0, "B" },
	{1116, 310, KEY_PAGEDOWN,     0, 0, "X" },
	{1116, 154, KEY_PAGEUP,       0, 0, "Y" },
	//
	{  54,  54, KEY_RIGHTMETA,    0, 0, "LT" },
	{  54,  74, KEY_RIGHTSHIFT,   0, 0, "LB" },
	{1192,  54, KEY_RIGHTALT,     0, 0, "RT" },
	{1192,  74, KEY_RIGHTCTRL,    0, 0, "RB" },
	{ 400, 354, BTN_RIGHT,        0, 0, "LNub" },
	{ 850, 354, BTN_LEFT,         0, 0, "RNub" },

	{ 614, 114, KEY_POWER,        0, 0, "POWER" },

	// top row
	{ 376, 114, KEY_ESC,          0, 0, "Esc" },
	{ 476, 114, KEY_PAUSE,        0, 0, "Usr1" },
	{ 790, 114, KEY_BRIGHTNESSUP, 0, 0, "*" },
	{ 882, 114, KEY_F11,          0, 0, "F11" },
	// center column
	{ 612, 178, KEY_LEFTALT,      0, 0, "Start" },
	{ 604, 248, KEY_LEFTCTRL,     0, 0, "Select" },
	{ 646, 314, KEY_MENU,         0, 0, "P" },
	// ABXY extension
	{1000, 168, KEY_INSERT,       0, 0, "I" },
	{ 996, 328, KEY_DELETE,       0, 0, "II" },
	// digits row
	{ 134, 418, KEY_1, 0, 0, "1" },
	{ 235, 418, KEY_2, 0, 0, "2" },
	{ 336, 418, KEY_3, 0, 0, "3" },
	{ 437, 418, KEY_4, 0, 0, "4" },
	{ 538, 418, KEY_5, 0, 0, "5" },
	{ 639, 418, KEY_6, 0, 0, "6" },
	{ 740, 418, KEY_7, 0, 0, "7" },
	{ 841, 418, KEY_8, 0, 0, "8" },
	{ 942, 418, KEY_9, 0, 0, "9" },
	{1043, 418, KEY_0, 0, 0, "0" },
	{1126, 418, KEY_BACKSPACE, 0, 0, "BS" },
	// TAB row
	{  60, 488, KEY_TAB,       0, 0, "Tab" },
	{ 184, 488, KEY_Q,         0, 0, "Q" },
	{ 286, 488, KEY_W,         0, 0, "W" },
	{ 388, 488, KEY_E,         0, 0, "E" },
	{ 490, 488, KEY_R,         0, 0, "R" },
	{ 592, 488, KEY_T,         0, 0, "T" },
	{ 694, 488, KEY_Y,         0, 0, "Y" },
	{ 796, 488, KEY_U,         0, 0, "U" },
	{ 898, 488, KEY_I,         0, 0, "I" },
	{1000, 488, KEY_O,         0, 0, "O" },
	{1102, 488, KEY_P,         0, 0, "P" },
	{1160, 488, KEY_ENTER,     0, 0, "Enter" },
	// middle letter row
	{ 134, 558, KEY_APOSTROPHE,0, 0, "\"" },
	{ 235, 558, KEY_A,         0, 0, "A" },
	{ 336, 558, KEY_S,         0, 0, "S" },
	{ 437, 558, KEY_D,         0, 0, "D" },
	{ 538, 558, KEY_F,         0, 0, "F" },
	{ 639, 558, KEY_G,         0, 0, "G" },
	{ 740, 558, KEY_H,         0, 0, "H" },
	{ 841, 558, KEY_J,         0, 0, "J" },
	{ 942, 558, KEY_K,         0, 0, "K" },
	{1043, 558, KEY_L,         0, 0, "L" },
	{1120, 558, KEY_LEFTSHIFT, 0, 0, "Shift" },
	// bottom letter row
	{  60, 638, KEY_COMMA,     0, 0, "," },
	{ 184, 638, KEY_DOT,       0, 0, "." },
	{ 286, 638, KEY_Z,         0, 0, "Z" },
	{ 388, 638, KEY_X,         0, 0, "X" },
	{ 490, 638, KEY_C,         0, 0, "C" },
	{ 592, 638, KEY_V,         0, 0, "V" },
	{ 694, 638, KEY_B,         0, 0, "B" },
	{ 796, 638, KEY_N,         0, 0, "N" },
	{ 898, 638, KEY_M,         0, 0, "M" },
	{1000, 638, KEY_SPACE,     0, 0, "SPACE" },
	{1198, 638, KEY_LEFTMETA,  0, 0, "Fn" },

	{  10, 324, BTN_TOUCH,     0, 0, "TOUCH" },
};

#define KI_COUNT (sizeof(key_items) / sizeof(key_items[0]))

static int screen_w;
static int screen_h;
static unsigned int stride;

static struct state {
	int nub_range_hit, all_btns_tested, lid_tested, ts_tested;
} st;
static int ts_old_x, ts_old_y;
static int lid_closed;

static void draw_ts_cross(unsigned int *fb, unsigned int col, int x, int y)
{
	int i, x1, y1;

	if (y >= 0 && y < screen_h)
		for (x1 = x - 5, i = 0; i < 11; i++, x1++)
			if (x1 >= 0 && x1 < screen_w) fb[y * stride/4 + x1] = col;
	
	if (x >= 0 && x < screen_w)
		for (y1 = y - 5, i = 0; i < 11; i++, y1++)
			if (y1 >= 0 && y1 < screen_h) fb[y1 * stride/4 + x] = col;

	if (x != ts_old_x && y != ts_old_y)
		st.ts_tested = 1;

	ts_old_x = x;
	ts_old_y = y;
}

static void text_out16(unsigned int *fb, int x, int y, unsigned int col, const char *text)
{
	int i,l;

	fb = fb + x + y*stride/4;

	for (i = 0; i < strlen(text); i++)
	{
		for (l=0;l<8;l++)
		{
			#define pix(fdmask,add) \
				if (fontdata8x8[((text[i])*8)+l]&fdmask) \
					fb[l*2*stride/4+add]=fb[l*2*stride/4+add+1]= \
					fb[(l*2+1)*stride/4+add]=fb[(l*2+1)*stride/4+add+1]=col
			pix(0x80,  0);
			pix(0x40,  2);
			pix(0x20,  4);
			pix(0x10,  6);
			pix(0x08,  8);
			pix(0x04, 10);
			pix(0x02, 12);
			pix(0x01, 14);
			#undef pix
		}
		fb += 8*2;
	}
}

static void text_out16_small(unsigned int *fb, int x, int y, unsigned int col, const char *text)
{
	int i,l;

	fb = fb + x + y*stride/4;

	for (i = 0; i < strlen(text); i++)
	{
		for (l=0;l<8;l++)
		{
			#define pix(fdmask,add) \
				if (fontdata8x8[((text[i])*8)+l]&fdmask) \
					fb[l*stride/4+add]=fb[l*stride/4+stride/4+add]=col
			pix(0x80,  0);
			pix(0x40,  1);
			pix(0x20,  2);
			pix(0x10,  3);
			pix(0x08,  4);
			pix(0x04,  5);
			pix(0x02,  6);
			pix(0x01,  7);
			#undef pix
		}
		fb += 8;
	}
}

static void redraw_nubs(unsigned int *fb, int x1, int y1, int x2, int y2)
{
	char buff[32];
	int fill = 0, y;

	fb += (256-64-8) * stride/4;

	if (st.nub_range_hit == 0xff && st.all_btns_tested && st.lid_tested && st.ts_tested)
		fill = 0x0f;

	// clear areas
	for (y = 0; y < (64+8*2)*2; y++) {
		memset(fb + stride/4*y + 412-64-8, fill, (64+8*2)*4*2);
		memset(fb + stride/4*y + 876-64-8, fill, (64+8*2)*4*2);
	}
	for (y = 130; y < 130+8; y++) {
		memset(fb + stride/4*y + 478, 0, 46*4*2);
		memset(fb + stride/4*y + 754, 0, 46*4*2);
	}

	text_out16(fb + 412-64-8, 64+8 + x1/4, 64+8 + y1/4, 0x0000ff, "@");
	text_out16(fb + 876-64-8, 64+8 + x2/4, 64+8 + y2/4, 0x0000ff, "@");

	if (x1 <= -NUB_RANGE_REQ) st.nub_range_hit |= 0x01;
	if (x1 >=  NUB_RANGE_REQ) st.nub_range_hit |= 0x02;
	if (y1 <= -NUB_RANGE_REQ) st.nub_range_hit |= 0x04;
	if (y1 >=  NUB_RANGE_REQ) st.nub_range_hit |= 0x08;
	if (x2 <= -NUB_RANGE_REQ) st.nub_range_hit |= 0x10;
	if (x2 >=  NUB_RANGE_REQ) st.nub_range_hit |= 0x20;
	if (y2 <= -NUB_RANGE_REQ) st.nub_range_hit |= 0x40;
	if (y2 >=  NUB_RANGE_REQ) st.nub_range_hit |= 0x80;

	snprintf(buff, sizeof(buff), "%4i", x1);
	text_out16_small(fb, 478, 130, (st.nub_range_hit & 0x03) == 0x03 ? COL_TESTED : COL_DEF, buff);
	snprintf(buff, sizeof(buff), "%4i", y1);
	text_out16_small(fb, 524, 130, (st.nub_range_hit & 0x0c) == 0x0c ? COL_TESTED : COL_DEF, buff);

	snprintf(buff, sizeof(buff), "%4i", x2);
	text_out16_small(fb, 754, 130, (st.nub_range_hit & 0x30) == 0x30 ? COL_TESTED : COL_DEF, buff);
	snprintf(buff, sizeof(buff), "%4i", y2);
	text_out16_small(fb, 800, 130, (st.nub_range_hit & 0xc0) == 0xc0 ? COL_TESTED : COL_DEF, buff);
}

static void redraw_keys_lid(unsigned int *fb)
{
	const key_item *key;
	int i;

	for (i = 0; i < KI_COUNT; i++)
	{
		key = &key_items[i];

		text_out16(fb, key->x, key->y,
			key->active ? COL_ACTIVE : (key->tested ? COL_TESTED : COL_DEF),
			key->name);
	}

	text_out16(fb, 10, 260, lid_closed ? COL_ACTIVE : (st.lid_tested ? COL_TESTED : COL_DEF), "LID");
}

static void set_key(int code, int val)
{
	key_item *key = NULL;
	int i;

	for (i = 0; i < KI_COUNT; i++)
	{
		if (key_items[i].code == code)
		{
			key = &key_items[i];
			break;
		}
	}

	if (key == NULL)
	{
		printf("%c unexpected key? (%i)\n", val ? '+' : '-', code);
	}
	else
	{
		key->active = !!val;
		key->tested |= !!val;
		printf("%c %s\n", val ? '+' : '-', key->name);
	}

	if (st.all_btns_tested)
		return;

	for (i = 0; i < KI_COUNT; i++)
	{
		if (key_items[i].code == KEY_MENU)
			continue;
		if (key_items[i].code == KEY_KPPLUS || key_items[i].code == KEY_KPMINUS)
			continue;
		if (!key_items[i].tested)
			break;
	}

	if (i == KI_COUNT)
		st.all_btns_tested = 1;
}

/* sound test */
#include <sys/soundcard.h>
#include <pthread.h>

static int snd_test_l, snd_test_r, snd_test_quit;
static pthread_cond_t snd_cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t snd_mutex = PTHREAD_MUTEX_INITIALIZER;

static void *sound_thread(void *arg)
{
	int i, ret, randfd, dspfd;
	int frag, bits, stereo, rate;
	short buf[8*1024];//2*22050 / 10];

	randfd = open("/dev/urandom", O_RDONLY);
	if (randfd == -1) {
		perror("open(\"/dev/urandom\")");
		return NULL;
	}

	dspfd = open("/dev/dsp", O_WRONLY);
	if (dspfd == -1) {
		perror("open(\"/dev/dsp\")");
		close(randfd);
		return NULL;
	}

	frag = (2 << 16) | 13;
	ret = ioctl(dspfd, SNDCTL_DSP_SETFRAGMENT, &frag);
	if (ret < 0)
		perror("SNDCTL_DSP_SETFRAGMENT");

	stereo = 1; bits = 16; rate = 22050;
	ret = ioctl(dspfd, SNDCTL_DSP_STEREO, &stereo);
	if (ret == 0)
		ret = ioctl(dspfd, SNDCTL_DSP_SETFMT, &bits);
	if (ret == 0)
		ret = ioctl(dspfd, SNDCTL_DSP_SPEED, &rate);
	if (ret < 0)
		perror("failed to set audio format");

	while (1) {
		pthread_mutex_lock(&snd_mutex);
		if (!snd_test_l && !snd_test_r)
			pthread_cond_wait(&snd_cond, &snd_mutex);
		pthread_mutex_unlock(&snd_mutex);

		if (snd_test_quit)
			break;

		ret = read(randfd, buf, sizeof(buf));
		if (ret == -1) {
			perror("urandom read");
			break;
		}

		if (!snd_test_l)
			for (i = 0; i < array_size(buf) / 2; i++)
				buf[i * 2] = 0;
		if (!snd_test_r)
			for (i = 0; i < array_size(buf) / 2; i++)
				buf[i * 2 + 1] = 0;

		ret = write(dspfd, buf, sizeof(buf));
		if (ret == -1) {
			perror("dsp write");
			break;
		}
	}

	close(randfd);
	close(dspfd);

	return NULL;
}

static void sound_init(void)
{
	pthread_t tid;
	int ret;

	ret = pthread_create(&tid, NULL, sound_thread, NULL);
	if (ret != 0) {
		fprintf(stderr, "pthread_create sound_thread: %d\n", ret);
		return;
	}
	pthread_detach(tid);
}

static void sound_do(int l, int r)
{
	pthread_mutex_lock(&snd_mutex);
	snd_test_l = l;
	snd_test_r = r;
	if (l || r || snd_test_quit)
		pthread_cond_signal(&snd_cond);
	pthread_mutex_unlock(&snd_mutex);
}

enum {
	DEV_PWRBTN,
	DEV_KEYPAD,
	DEV_BUTTONS1,
	DEV_BUTTONS2,
	DEV_TS,
	DEV_LNUB,
	DEV_RNUB,
	DEVS_TOTAL
};

int main(int argc, char *argv[])
{
	unsigned int *screen;
	struct state oldstate;
	int fbdev, ifd[DEVS_TOTAL];
	int i, id, imaxfd = 0, ts_x = 0, ts_y = 0;
	int nubx[2] = {0,0}, nuby[2] = {0,0};
	int pressed_l = 0, pressed_r = 0;
	struct fb_fix_screeninfo fbfix;
	struct fb_var_screeninfo fbvar;
	int nubs_warned = 0;
	int ret, retval = 0;

	memset(ifd, 0xff, sizeof(ifd));

	fbdev = open(argv[1] ?: "/dev/fb0", O_RDWR);
	if (fbdev == -1)
	{
		perror("open(fb) failed");
		return 1;
	}

	ret = ioctl(fbdev, FBIOGET_FSCREENINFO, &fbfix);
	if (ret == -1)
	{
		perror("ioctl(FBIOGET_FSCREENINFO)");
		return 1;
	}

	printf("id: \"%s\"\n", fbfix.id);
	printf("line bytes: %i\n", fbfix.line_length);
	stride = fbfix.line_length;

	ret = ioctl(fbdev, FBIOGET_VSCREENINFO, &fbvar);
	if (ret == -1)
	{
		perror("ioctl(FBIOGET_VSCREENINFO)");
		return 1;
	}

	printf("%dx%d (%dx%d) @%d\n", fbvar.xres, fbvar.yres,
		fbvar.xres_virtual, fbvar.yres_virtual, fbvar.bits_per_pixel);
	screen_w = fbvar.xres;
	screen_h = fbvar.yres;

	screen = mmap(0, stride*screen_h, PROT_WRITE|PROT_READ, MAP_SHARED, fbdev, 0);
	if (screen == MAP_FAILED)
	{
		perror("mmap");
		return 1;
	}

	memset(screen, 0, stride*screen_h);

	sound_init();

	for (id = 0; ; id++)
	{
		char fname[64];
		char name[256] = { 0, };
		int fd;

		snprintf(fname, sizeof(fname), "/dev/input/event%i", id);
		fd = open(fname, O_RDONLY);
		if (fd == -1)
		{
			break;
		}

		ioctl(fd, EVIOCGNAME(sizeof(name)), name);

		/*
		/dev/input/event0:      user-button
		/dev/input/event1:      pyra-gpio-keys@1
		/dev/input/event2:      pyra-gpio-keys@2
		/dev/input/event3:      palmas_pwron
		/dev/input/event4:      tca8418
		/dev/input/event5:      nub0
		/dev/input/event6:      nub1
		/dev/input/event7:      TSC2007 Touchscreen
		*/

		if      (strcasestr(name, "power") != NULL
			 || strcasestr(name, "pwron") != NULL)
			ifd[DEV_PWRBTN] = fd;
		else if (strcasestr(name, "keypad") != NULL
			 || strcasestr(name, "tca8418") != NULL)
			ifd[DEV_KEYPAD] = fd;
		else if (strcmp(name, "user-buttons") == 0)
			ifd[DEV_BUTTONS1] = fd;
		else if (strcmp(name, "pyra-game-buttons") == 0)
			ifd[DEV_BUTTONS2] = fd;
		else if (strcasestr(name, "touchscreen") != NULL)
			ifd[DEV_TS] = fd;
		else if (strcmp(name, "nub0") == 0)
			ifd[DEV_LNUB] = fd;
		else if (strcmp(name, "nub1") == 0)
			ifd[DEV_RNUB] = fd;
		else
		{
			printf("skipping \"%s\"\n", name);
			close(fd);
			continue;
		}
		if (imaxfd < fd) imaxfd = fd;
	}

	if (ifd[DEV_PWRBTN]  == -1) printf("Warning: couldn't find pwrbutton device\n");
	if (ifd[DEV_KEYPAD]  == -1) printf("Warning: couldn't find keypad device\n");
	if (ifd[DEV_TS]      == -1) printf("Warning: couldn't find touchscreen device\n");
	if (ifd[DEV_LNUB]    == -1) printf("Warning: couldn't find nub1 device\n");
	if (ifd[DEV_RNUB]    == -1) printf("Warning: couldn't find nub2 device\n");
	if (ifd[DEV_BUTTONS1] == -1) {
		printf("Error: couldn't find button device\n");
		/* buttons needed to exit this, so bail out */
		retval = -1;
		goto end;
	}

#ifdef VALIDATION
	int early_leave_attempt = 0;
	text_out16_small(screen, 220, 3, COL_DEF,
			 " press menu/pyra button to exit");
#else
	text_out16_small(screen, 320, 3, COL_DEF, "Press L+R to exit");
#endif
	memset(&oldstate, 0xff, sizeof(oldstate));

#ifdef VALIDATION
	while (1)
#else
	while (!pressed_l || !pressed_r)
#endif
	{
		struct input_event ev[64];
		int fd = -1, rd, which, ret;
		fd_set fdset;

#ifdef VALIDATION
		if (early_leave_attempt > 1) {
			char buff[128];
			if (memcmp(&st, &oldstate, sizeof(st))) {
				memset(screen, 0, stride * 12 * 4);
				memcpy(&oldstate, &st, sizeof(st));
			}
			snprintf(buff, sizeof(buff),
				"not everything tested (%s%s%s%s), LB+RB to override",
				st.nub_range_hit == 0xff ? "" : "nubs",
				st.all_btns_tested ? "" : " buttons",
				st.lid_tested ? "" : " lid_switch",
				st.ts_tested ? "" : " touchscreen");
			if (st.nub_range_hit == 0xff && st.all_btns_tested && st.lid_tested && st.ts_tested)
				text_out16(screen, 220, 3, 0x007c00,
					"press menu/pyra button to exit");
			else
				text_out16_small(screen, 38, 3, 0xf80000, buff);
		}
#endif
		draw_ts_cross(screen, 0x000000, ts_old_x, ts_old_y);
		redraw_keys_lid(screen);
		redraw_nubs(screen, nubx[0], nuby[0], nubx[1], nuby[1]);
		draw_ts_cross(screen, 0xff0000, ts_x, ts_y);

		FD_ZERO(&fdset);
		for (i = 0; i < DEVS_TOTAL; i++)
			if (ifd[i] != -1)
				FD_SET(ifd[i], &fdset);

		ret = select(imaxfd + 1, &fdset, NULL, NULL, NULL);
		if (ret == -1)
		{
			perror("select");
			break;
		}

		fd = -1;
		for (i = 0; i < DEVS_TOTAL; i++) {
			if (ifd[i] != -1 && FD_ISSET(ifd[i], &fdset)) {
				fd = ifd[i];
				break;
			}
		}

		rd = read(fd, ev, sizeof(ev));
		if (rd < (int) sizeof(ev[0])) {
			perror("\nevtest: error reading");
			break;
		}

		for (i = 0; i < rd / sizeof(ev[0]); i++)
		{
			switch (ev[i].type) {
			case EV_SYN:
				break;
			case EV_KEY:
				set_key(ev[i].code, ev[i].value);

				if (ev[i].code == KEY_RIGHTSHIFT)
					pressed_l = !!ev[i].value;
				if (ev[i].code == KEY_RIGHTCTRL)
					pressed_r = !!ev[i].value;
#ifdef VALIDATION
				if (ev[i].code == KEY_MENU && !!ev[i].value) {
					if (st.nub_range_hit == 0xff && st.all_btns_tested && st.lid_tested && st.ts_tested)
						goto end;
					if (pressed_l && pressed_r) {
						retval = 1;
						goto end;
					}
					early_leave_attempt++;
				}
#endif
				if (ev[i].code == KEY_LEFT)
					sound_do(!!ev[i].value, snd_test_r);
				if (ev[i].code == KEY_RIGHT)
					sound_do(snd_test_l, !!ev[i].value);
				break;
			case EV_ABS:
				if (fd == ifd[DEV_LNUB] || fd == ifd[DEV_RNUB]) {
					which = (fd == ifd[DEV_LNUB]) ? 0 : 1;
					switch (ev[i].code) {
					case ABS_X:
						nubx[which] = ev[i].value;
						break;
					case ABS_Y:
						nuby[which] = ev[i].value;
						break;
					default:
						printf("unexpected nub EV_ABS code: %i\n",
							ev[i].code);
						break;
					}
				}
				else if (fd == ifd[DEV_TS]) {
					// just a guess, my ts is destroyed
					switch (ev[i].code) {
					case ABS_X:
						ts_x = ev[i].value * 1280 / 720;
						break;
					case ABS_Y:
						ts_y = ev[i].value * 720 / 1280;
						break;
					case ABS_PRESSURE:
						break;
					default:
						printf("unexpected nub EV_ABS code: %i\n",
							ev[i].code);
						break;
					}
				}
				break;
			case EV_SW:
				if (ev[i].code == SW_LID) {
					lid_closed = ev[i].value;
					st.lid_tested |= ev[i].value;
				}
				else
					printf("unexpected EV_SW code: %i\n", ev[i].code);
				break;
			case EV_REL:
				if (!nubs_warned) {
					fprintf(stderr, "please switch nubs to absolute mode!\n");
					nubs_warned = 1;
				}
				break;
			case EV_MSC:
				if (ev[i].code == MSC_SCAN)
					break;
				/* fallthrough */
			default:
				printf("unexpected event: type %i, code %d\n", ev[i].type, ev[i].code);
				break;
			}
		}
	}

	snd_test_quit = 1;
	sound_do(0, 0);

end:
	for (i = 0; i < DEVS_TOTAL; i++)
		if (i != DEV_TS && ifd[i] != -1)
			close(ifd[i]);
	memset(screen, 0, stride*screen_h);
	munmap(screen, stride*screen_h);
	close(fbdev);
	return retval;
}

